#!/usr/bin/perl
# update_map.pl - update maps to reflect subtopic structures in topic files
#   v0.1  chrispy  


# Prerequisites:
#  sudo apt-get install make cpanminus  (makes it much easier to install Perl modules)
#  sudo cpanm install XML::Twig utf8::all

use strict;
use warnings;

use XML::Twig;
use utf8::all;
#use Acme::Tools;
use File::Find;
use File::Spec;
use File::Basename;

# analyze all map files and remember the dita file references
my %maps_of_dita_files = ();
my %topicrefs_of_dita_files = ();
my %map_twigs = ();
my %keys_in_map = ();
{
 my @map_files = ();
# find({ wanted => sub { push @map_files, $File::Find::name if (m!\.ditamap$!); }, follow => 1 }, '.');  # get .ditamap files recursively down the directories
 find({ wanted => sub { push @map_files, $File::Find::name if (m!\.ditamap$!); }, follow => 0 }, './dita');  # get .ditamap files from a single path ('./dita')
 foreach my $map_file (map {File::Spec->rel2abs($_)} @map_files) {
  my $these_keys = ($keys_in_map{$map_file} = {});
  $map_twigs{File::Spec->rel2abs($map_file)} = my $this_twig = XML::Twig->new(
   keep_spaces => 0,
   twig_handlers => {
    '*[@keys]' => sub { $these_keys->{$_->att('keys')} = 1; return 1; },  # store key values to avoid collisions later
    'topicref' => sub {
     return 1 if $_->att('href') =~ m!#!;  # only interested in subtopics, not top-level topics
     return 1 if $_->att('href') =~ m!\.ditamap$!;  # topicref shouldn't point to a map anyway...
     my $dita_file = File::Spec->rel2abs($_->att('href'), dirname($map_file));
     push @{$maps_of_dita_files{$dita_file}}, $map_file;
     push @{$topicrefs_of_dita_files{$dita_file}}, $_;
    },
   })->parsefile($map_file);
  $this_twig->root->set_att('#keys', $these_keys);
 }
}


# make a list of DITA files to check - default is to check them all
my @dita_files = @ARGV;
@dita_files = (sort keys %maps_of_dita_files) if !@dita_files;

# compare each DITA topic against its topicref
foreach my $dita_file (map {File::Spec->rel2abs($_)} @dita_files) {
print "FILE:'$dita_file'\n";
 do {print "Error: No maps found for '$dita_file'.\n"; next;} if !defined($topicrefs_of_dita_files{$dita_file});
 do {print "Error: DITA file '$dita_file' not found.\n"; next;} if !(-e $dita_file);

 my $dita_root = XML::Twig->new(
  keep_spaces => 0,
  ignore_elts => {abstract => 'discard', prolog => 'discard', body => 'discard', 'related-links' => 'discard'},
  comments => 'drop', pi => 'drop',
 )->parsefile($dita_file)->root;
 foreach my $topicref (@{$topicrefs_of_dita_files{$dita_file}}) {
  if (compare_topicref_to_topic($topicref, $dita_root) == 0) {
   $topicref->root->set_att('#changed', 1);
   make_topicref_from_topic($topicref, $dita_root);
  }
 }
}

foreach my $map_file (grep {$map_twigs{$_}->root->att('#changed')} sort keys %map_twigs) {
 write_entire_file("${map_file}", $map_twigs{$map_file}->sprint(pretty_print => 'indented'));
 my $r = File::Spec->abs2rel($map_file);
 print "Updated '$map_file'.\n";
}



sub subtopic_id {
 my $href = shift;
 return $href =~ s!^.*#!!r;
}

sub simple_name {
 my $elt = shift or return undef;
 my $title = $elt->att('#title');
 if ($title->first_descendant('*[@keyref]')) {
#  $_->set_tag('#PCDATA')->set_text($xmlvars{$_->att('keyref')}) for $elt->descendants('*[@keyref]');
  $_->set_tag('#PCDATA')->set_text($_->att('keyref')) for $title->descendants('*[@keyref]');
 }
 my $name = lc $title->text;
 $name =~ s!\b(a|the|and|or)\b!!gs;  # TO-DO - don't simplify a section with 'A' by itself (alphabetical names by letter)
 $name =~ s!\W+!_!gs;
 $name =~ s!(^_+|_+$)!!gs;
 return $name;
}

sub compare_topicref_to_topic {
 my ($tr, $t) = @_;
 my @subrefs = $tr->children('topicref[@href =~ /#/]');
 my @subtopics = grep {!$_->matches('title')} $t->children;
 return 0 if scalar(@subrefs) != scalar(@subtopics);
 for (my $i = 0; $i < @subtopics; $i++) {
  my ($subref, $subtopic) = ($subrefs[$i], $subtopics[$i]);
  return 0 if (subtopic_id($subref->att('href'))) ne $subtopic->id;  # check ID value 
  return 0 if ($subref->att('toc') or 'yes') ne ($subtopic->matches('module') ? 'no' : 'yes');
  return 0 if compare_topicref_to_topic($subref, $subtopic) == 0;  # check children (recursive)
 }
 return 1; 
}

sub make_topicref_from_topic {
 my ($tr, $t) = @_;
 my $map = $tr->root;
 my $topic_href = $tr->att('href');
 my $existing_topicrefs = XML::Twig::Elt->new('topicrefs', $tr->cut_children('topicref[@href =~ /#/]'));
 my $new_topicrefs = XML::Twig::Elt->new('topicrefs', $t->children_copy);

 delete $map->att('#keys')->{$_} for map {$_->att('keys')} $existing_topicrefs->descendants('*[@keys]');  # pull these out of this map's key space
 my %topicref_for_id = map {subtopic_id($_->att('href')) => $_} $existing_topicrefs->children;

 $_->parent->set_att('#title', $_->cut) for $new_topicrefs->descendants('title');
 foreach my $this_tr ($new_topicrefs->descendants) {
  my $id = $this_tr->att('id');
  $this_tr->set_att('toc', 'no') if $this_tr->matches('module');
  $this_tr->set_tag('topicref');
  $this_tr->set_att('href', "${topic_href}#${id}")->del_att('id');
  my $key = defined($topicref_for_id{$id}) ? $topicref_for_id{$id}->att('keys') : simple_name($this_tr);
  while (defined($map->att('#keys')->{$key})) {$key = "XX__${key}__XX";}
  $this_tr->set_att('keys', $key);
  $map->att('#keys')->{$key} = 1;  # add back into this map's key space
  $this_tr->set_empty if !$this_tr->has_children;
 }
 $new_topicrefs->paste('first_child', $tr)->erase;
}

sub write_entire_file {
 my ($filename, $contents) = @_;
 $contents =~ s!\n?$!\n!s;  # add LF if needed
 open(FILE, ">$filename") or die "can't open $filename for write: $!";
 binmode(FILE);  # don't convert LFs to CR/LF on Windows
 binmode(FILE, ":encoding(utf-8)");  # the UTF-8 package checks and enforces this
 print FILE $contents;
 close FILE;
}
